-- Rider Bump
-- removes mid-air playerbumps and nerfs ground-based ones
-- you're free to edit and use this code for whatever; most of this is already taken from the game's source

local bumptime = (TICRATE*100)/583; 
local wipeoutslowtime = 20;

local function getKartSpeed(player, doboostpower)
	local k_speed = 150;
	local g_cc = FRACUNIT;
	local xspd = 3072;		-- 4.6875 aka 3/64
	local kartspeed = player.kartspeed;
	local finalspeed;

	if (doboostpower and (not player.kartstuff[k_pogospring]) and (not P_IsObjectOnGround(player.mo)))
		return (75*mapobjectscale); -- air speed cap
	end

	--switch (gamespeed) -- lua doesn't have switch statements
	if (gamespeed == 0)
		g_cc = 53248 + xspd; --  50cc =  81.25 + 4.69 =  85.94%
	elseif (gamespeed == 2)
		g_cc = 77824 + xspd; -- 150cc = 118.75 + 4.69 = 123.44%
	elseif (gamespeed == 1)
		g_cc = 65536 + xspd; -- 100cc = 100.00 + 4.69 = 104.69%
	end

	if (G_BattleGametype() and (player.kartstuff[k_bumper] <= 0))
		kartspeed = 1;
	end

	k_speed = ($ + kartspeed*3); -- 153 - 177

	local pScale = mapobjectscale
	if (player.mo and player.mo.valid) then pScale = player.mo.scale end
	finalspeed = FixedMul(FixedMul(k_speed<<14, g_cc), pScale)

	if (doboostpower)
		return FixedMul(finalspeed, player.kartstuff[k_boostpower]+player.kartstuff[k_speedboost])
	end
	return finalspeed;
end

local function K_GetMobjWeight(mobj, against, nerfVal)
	local weight = 5<<FRACBITS;

	if ((mobj.type) == MT_PLAYER)

		if (not mobj.player) then return; end

		if (against.player and not against.player.kartstuff[k_spinouttimer] and mobj.player.kartstuff[k_spinouttimer])
			weight = 0; -- Do not bump
		else
			weight = (mobj.player.kartweight)<<FRACBITS;
			if (mobj.player.speed > getKartSpeed(mobj.player, false))
				weight = ($ + (mobj.player.speed - getKartSpeed(mobj.player, false))/8);
            end

        end

	elseif ((mobj.type) == MT_FALLINGROCK)

		if (against.player)
			if (against.player.kartstuff[k_invincibilitytimer]
				or against.player.kartstuff[k_growshrinktimer] > 0)
				weight = 0;
			else
				weight = (against.player.kartweight)<<FRACBITS;
            end
		end
            
	elseif (((mobj.type) == MT_ORBINAUT) or ((mobj.type) == MT_ORBINAUT_SHIELD)) -- orbis

		if (against.player)
			weight = (against.player.kartweight)<<FRACBITS;
        end

	elseif (((mobj.type) == MT_JAWZ) or ((mobj.type) == MT_JAWZ_DUD) or ((mobj.type) == MT_JAWZ_SHIELD))

		if (against.player)
			weight = (against.player.kartweight+3)<<FRACBITS;
		else
			weight = 8<<FRACBITS;
        end

	end

    if (nerfVal)
        weight = ($ / abs(nerfVal));
    end

	return weight;
end


local function KP_KartBouncing(mobj1, mobj2, bounce, solid, nerf)
	local fx; -- mobj_t
	local momdifx, momdify; -- fixed_t
	local distx, disty; -- fixed_t
	local dot, force; -- fixed_t
	local mass1, mass2; -- fixed_t

	if ( (not (mobj1 and mobj1.valid)) or (not (mobj2 and mobj2.valid)))
		return;
    end

	-- Don't bump when you're being reborn
	if ((mobj1.player and mobj1.player.playerstate != PST_LIVE)
		or (mobj2.player and mobj2.player.playerstate != PST_LIVE))
		return;
    end

	if ((mobj1.player and mobj1.player.kartstuff[k_respawn])
		or (mobj2.player and mobj2.player.kartstuff[k_respawn]))
		return;
    end

	-- Don't bump if you're flashing
	local flash;

	if (mobj1.player) then flash = K_GetKartFlashing(mobj1.player); end
	if (mobj1.player and mobj1.player.powers[pw_flashing] > 0 and mobj1.player.powers[pw_flashing] < flash)
		if (mobj1.player.powers[pw_flashing] < flash-1)
			mobj1.player.powers[pw_flashing] = ($ + 1);
        end

		return;
    end


	if (mobj2.player) then flash = K_GetKartFlashing(mobj2.player); end -- WHY IS THERE A NILCHECK HERE
	if (mobj2.player and mobj2.player.powers[pw_flashing] > 0 and mobj2.player.powers[pw_flashing] < flash)
		if (mobj2.player.powers[pw_flashing] < flash-1)
			mobj2.player.powers[pw_flashing] = ($ + 1);
        end

		return;
    end

	-- Don't bump if you've recently bumped
	if (mobj1.player and mobj1.player.phs_justbumped)
		mobj1.player.phs_justbumped = bumptime;

		return;
    end

	if (mobj2.player and mobj2.player.phs_justbumped)
		mobj2.player.phs_justbumped = bumptime;

		return;
    end

	mass1 = K_GetMobjWeight(mobj1, mobj2, 1);

	if (solid == true and mass1 > 0)
		mass2 = mass1;
	else
		mass2 = K_GetMobjWeight(mobj2, mobj1, 1);
    end

	momdifx = mobj1.momx - mobj2.momx;
	momdify = mobj1.momy - mobj2.momy;

	-- Adds the OTHER player's momentum times a bunch, for the best chance of getting the correct direction
	distx = (mobj1.x + mobj2.momx*3) - (mobj2.x + mobj1.momx*3);
	disty = (mobj1.y + mobj2.momy*3) - (mobj2.y + mobj1.momy*3);

	if (distx == 0 and disty == 0) then return; end -- if there's no distance between the 2, they're directly on top of each other, don't run this

	-- Normalize distance to the sum of the two objects' radii, since in a perfect world that would be the distance at the point of collision...
	local dist = P_AproxDistance(distx, disty);
	local nx = FixedDiv(distx, dist);
	local ny = FixedDiv(disty, dist);

	dist = (dist and dist or 1);
	
	local nerfDiv = (nerf and nerf or 1);
	
	distx = FixedMul(mobj1.radius + mobj2.radius, nx);
	disty = FixedMul(mobj1.radius + mobj2.radius, ny);

	if (momdifx == 0 or momdify == 0)
		-- If there's no momentum difference, they're moving at exactly the same rate. Pretend they moved into each other.
		momdifx = -nx;
		momdify = -ny;
    end

	-- if the speed difference is less than this let's assume they're going proportionately faster from each other
	if (P_AproxDistance(momdifx, momdify) < (25*mapobjectscale))
		local momdiflength = P_AproxDistance(momdifx, momdify);
		local normalisedx = FixedDiv(momdifx, momdiflength);
		local normalisedy = FixedDiv(momdify, momdiflength);
		momdifx = FixedMul((25*mapobjectscale), normalisedx);
		momdify = FixedMul((25*mapobjectscale), normalisedy);
    end

	dot = FixedMul(momdifx, distx) + FixedMul(momdify, disty);

	if (dot >= 0) then return; end -- They're moving away from each other

	force = FixedDiv(dot, FixedMul(distx, distx)+FixedMul(disty, disty));

	if ((bounce == true) or mass2 > 0) -- Perform a Goomba Bounce.
		mobj1.momz = -mobj1.momz;
	else
		local newz = mobj1.momz;

		if (mass2 > 0) then mobj1.momz = mobj2.momz; end

		if ((mass1 > 0) and solid == false) then mobj2.momz = newz; end
    end
	
	local a = mobj1.angle;

	local b = mobj2.angle - a;
	
	local bonp = false;
	
	//print((abs(b)/ANG1))
	
	if ((abs(b)/ANG1) <= 120)
		bonp = true;
	end	
	
	/*if (bonp)
		print("bonnnp")
	end*/

	if (mass2 > 0)
		if (bonp == false)
			mobj1.momx = (mobj1.momx - (FixedMul(FixedMul(FixedDiv(2*mass2, mass1 + mass2), force), distx)/nerfDiv));
			mobj1.momy = (mobj1.momy - (FixedMul(FixedMul(FixedDiv(2*mass2, mass1 + mass2), force), disty)/nerfDiv));
		else -- P1 gets the bump instead, don't fuck with P2
			mobj2.momx = (mobj2.momx - (FixedMul(FixedMul(FixedDiv(2*mass1, mass1 + mass2), force), -distx)/nerfDiv));
			mobj2.momy = (mobj2.momy - (FixedMul(FixedMul(FixedDiv(2*mass1, mass1 + mass2), force), -disty)/nerfDiv));
		end
    end

	if ((mass1 > 0) and solid == false and (bonp == false))
		mobj2.momx = (mobj2.momx - (FixedMul(FixedMul(FixedDiv(2*mass1, mass1 + mass2), force), -distx)/nerfDiv));
		mobj2.momy = (mobj2.momy - (FixedMul(FixedMul(FixedDiv(2*mass1, mass1 + mass2), force), -disty)/nerfDiv));
    end

	-- Do the bump fx when we've CONFIRMED we can bump.
	S_StartSound(mobj1, sfx_s3k49);

	fx = P_SpawnMobj(mobj1.x/2 + mobj2.x/2, mobj1.y/2 + mobj2.y/2, mobj1.z/2 + mobj2.z/2, MT_BUMP);

	if (mobj1.eflags & MFE_VERTICALFLIP)
		fx.eflags = ($ | MFE_VERTICALFLIP);
	else
		fx.eflags = ($ & ~MFE_VERTICALFLIP);
    end

	P_SetScale(fx, mobj1.scale);

	-- Because this is done during collision now, rmomx and rmomy need to be recalculated
	-- so that friction doesn't immediately decide to stop the player if they're at a standstill
	-- Also set justbumped here
	if (mobj1.player)
		mobj1.player.rmomx = mobj1.momx - mobj1.player.cmomx;
		mobj1.player.rmomy = mobj1.momy - mobj1.player.cmomy;
		mobj1.player.phs_justbumped = bumptime;
		if (mobj1.player.kartstuff[k_spinouttimer])
			mobj1.player.kartstuff[k_wipeoutslow] = wipeoutslowtime+1;
			mobj1.player.kartstuff[k_spinouttimer] = max(wipeoutslowtime+1, mobj1.player.kartstuff[k_spinouttimer]);
        end
		if (mobj1.player.phs_driftsave)
			-- prevent wrongdrifts
			mobj1.player.kartstuff[k_driftend] = 0;
			mobj1.player.kartstuff[k_driftboost] = 0;
			mobj1.player.kartstuff[k_drift] = mobj1.player.phs_driftsave;
		end
	end

	if (mobj2.player)
		mobj2.player.rmomx = mobj2.momx - mobj2.player.cmomx;
		mobj2.player.rmomy = mobj2.momy - mobj2.player.cmomy;
		mobj2.player.phs_justbumped = bumptime;
		if (mobj2.player.kartstuff[k_spinouttimer])
			mobj2.player.kartstuff[k_wipeoutslow] = wipeoutslowtime+1;
			mobj2.player.kartstuff[k_spinouttimer] = max(wipeoutslowtime+1, mobj2.player.kartstuff[k_spinouttimer]);
        end

		if (mobj2.player.phs_driftsave)
			-- prevent wrongdrifts
			mobj2.player.kartstuff[k_driftend] = 0;
			mobj2.player.kartstuff[k_driftboost] = 0;
			mobj2.player.kartstuff[k_drift] = mobj2.player.phs_driftsave;
		end
	end
end

local function isBumpable(thing, p)
    local result = false;
    local tmthing;

    if (p)
        tmthing = p.mo;
    end

    if (thing.type == MT_SMK_ICEBLOCK)
    or (thing.type == MT_PLAYER)
    or (thing.type == MT_FALLINGROCK)
    or (thing.type == MT_BLUEROBRA_HEAD) 
    or (thing.type == MT_BLUEROBRA_JOINT)
    or (thing.type == MT_SMK_PIPE)
        result = true;
    end

    if (tmthing)
        if ((thing.type == MT_SMK_THWOMP) and (not (tmthing.z < thing.z and thing.momz < 0)))
        or ((thing.flags & MF_SOLID) and (P_IsObjectOnGround(thing) and (tmthing.momz < 0) ))
            result = true;
        end
    end

    return result;
end

local function nitroAssert() -- if you include the custom bumpcode as a standalone it's best you remove this
	if (driftmodData.isOn ~= true) then return false; end
	if (driftmodData.bumpCode ~= true) then return false; end
	
	return true;
end

local function KartRiderBumpCode(mo, tm)
	
	if (not nitroAssert()) then return; end
	
    -- see if it went over / under
    local zCheck = false;
    local collideP;

    local p = mo.player;

    
	if (tm.z > mo.z + tm.height) then zCheck = true; end -- overhead
	if (tm.z + tm.height < mo.z) then zCheck = true; end -- underneath

	if (zCheck == true) then return; end -- I COMPLETELY misinterpereted this

    if (isBumpable(tm, p) and (P_IsObjectOnGround(mo) and P_IsObjectOnGround(tm))) -- this should only run if both players are on the ground
        KP_KartBouncing(mo, tm, false, true, 12);
    end
end


addHook("MobjThinker", function(mo)
    local p = mo.player;

	if (not nitroAssert()) then return; end

    if (p.phs_hititem)
        p.phs_hititem = ($ - 1);
    else
        p.kartstuff[k_justbumped] = bumptime; -- constantly reset justbumped
    end

    if (p.phs_justbumped)
        p.phs_justbumped = ($ - 1);
	else
		p.phs_driftsave = p.kartstuff[k_drift]; -- only store if we haven't bumped again
	end

    if (not P_IsObjectOnGround(mo)) -- do leeway if you're in midair
        p.phs_justbumped = (TICRATE*3/4);
    end

end, MT_PLAYER)

addHook("MobjCollide", KartRiderBumpCode, MT_PLAYER)

addHook("MobjMoveCollide", KartRiderBumpCode, MT_PLAYER)

addHook("ShouldSpin", function(p, i, src)
    if (i)
        if (i.type == MT_ORBINAUT_SHIELD)
        or (i.type == MT_JAWZ_SHIELD)
        or (i.type == MT_ORBINAUT) 
        or (i.type == MT_JAWZ) 
        or (i.type == MT_JAWZ_DUD)
            p.kartstuff[k_justbumped] = 0; -- hit an item
            p.phs_hititem = bumptime;
        end
    end
end)